#!/usr/bin/env python3

""" Strong typing. """

__all__ = [
    'to_vector',
    'to_array',
    'to_list_of_type',
    'is_jax_vector',
    'is_jax_array',
    'jax_array_type',
    'least_base_type',
]

import jax
import jaxlib
import jax.numpy as jnp

import types
import typing
from _collections_abc import dict_values

jax_array_type = jax.Array
jax_tracer_type = jax.core.Tracer

def to_vector(arg):
    if isinstance(arg, jax_tracer_type):
        return arg # let jax trace
    if isinstance(arg, jax_array_type):
        assert arg.ndim == 1
        return arg
    if hasattr(arg, '__array__'):
        arg = arg.__array__()
        assert arg.ndim == 1
        return arg
    raise TypeError(f"Cannot coerce to vector: {arg!r}")


def to_array(arg):
    if isinstance(arg, jax_tracer_type):
        return arg # let jax trace
    if isinstance(arg, jax_array_type):
        assert arg.ndim == 2
        return arg
    if isinstance(arg, dict):
        arg = list(arg.values())
    if isinstance(arg, dict_values):
        arg = list(arg)
    if isinstance(arg, (list, tuple, set)):
        vectors = [to_vector(o) for o in arg]
        return jnp.stack(vectors, axis=0)
    if hasattr(arg, '__array__'):
        arg = arg.__array__()
        assert arg.ndim == 2
        return arg
    raise TypeError(f"Cannot coerce to array: {arg!r}")


def to_list_of_type(arg, type):
    if isinstance(arg, type):
        return [arg]
    elif is_instance(arg, list[type]):
        return arg
    elif is_instance(arg, tuple[type]):
        return list(arg)
    else:
        clsname = type.__name__
        raise TypeError(f"Cannot coerce to list of {clsname}: {arg!r}")

def is_jax_array(arg):
    return isinstance(arg, jax_array_type)

def is_jax_vector(arg):
    return isinstance(arg, jax_array_type) and arg.ndim == 1

def least_base_type(*types):
    from functools import reduce
    from operator import and_
    from collections import Counter
    return next(iter(reduce(and_, (Counter(t.mro()) for t in types))))

to_thought = to_vector # compatability
__all__.append('to_thought')




#!/usr/bin/env python3

import pytest

from think import Object, Type
from think import Bool, Int, Str, Float
from think import BoolType, StrType, IntType, FloatType


# derived meta
class Boolean(BoolType): pass
class String(StrType): pass
class Integer(IntType): pass
class Floating(FloatType): pass
class StrType1(StrType): pass
class StrType2(StrType): pass


def test_derived_meta_object_attribute():
    assert Boolean.object  is type
    assert String.object   is type
    assert Integer.object  is type
    assert Floating.object is type


def test_derived_meta_base_attribute():
    assert Boolean.base  is Bool
    assert String.base   is Str
    assert Integer.base  is Int
    assert Floating.base is Float


def test_derived_meta_not_equal_implies_types_not_equal():
    assert StrType1('Ticker') != StrType2('Ticker')


def test_derived_meta_not_equal_implies_objects_not_equal():
    assert StrType1('Ticker')('SPY') != StrType2('Ticker')('SPY')



# ways of requesting the bases that our new class will subclass

class Ticker(Str):
    pass

class TickerType1(Type):
    base = Ticker

class TickerType2(Type):
    def __new__(cls, name):
        return Type.__new__(cls, name, Ticker)

def test_derived_meta_returns_subclasses_of_base_if_passed_in_metaclass_body():
    ETF = TickerType1('ETF')
    assert issubclass(ETF, Ticker)


def test_derived_meta_returns_subclasses_of_base_if_passed_in_metaclass_new():
    ETF = TickerType2('ETF')
    assert issubclass(ETF, Ticker)


if False:
    # obscure __classcell__ error occurring
    # when we make a class Char(Str): pass
    # *twice* (but not once), and this is
    # happening iff we cache types, so don't
    # cache types for now.
    def test_idempotence_for_derived_meta():
        assert Boolean('Alive') == Boolean('Alive')
        assert String('Name')   == String('Name')
        assert Integer('Age')   == Integer('Age')
        assert Floating('Size') == Floating('Size')





#!/usr/bin/env python3

__all__ = [
    'Thought',
    'thought_dim',
    'random_normal',
    'new_thought',
    'new_thoughts',
    'ordinal_basis',
]

import jax.numpy as jnp

from think.internals import State

STATE = State()

THOUGHT_DIM = 1024

def random_normal(shape):
    return STATE.normal(shape)

def new_thought():
    denominator = jnp.sqrt(THOUGHT_DIM)
    return STATE.normal([THOUGHT_DIM])/denominator

def new_thoughts(n):
    denominator = jnp.sqrt(THOUGHT_DIM)
    return STATE.normal([n, THOUGHT_DIM])/denominator

def ordinal_basis(n, m=1):
    ortho = [new_thought() for k in range(n+1)]
    basis = []
    thetas = jnp.linspace(0, jnp.pi/2, m+1)[+1:-1]
    for a,c in zip(ortho[:-1], ortho[+1:]):
        bs = [a*jnp.cos(theta) + c*jnp.sin(theta) for theta in thetas]
        basis += [a,*bs]
    return jnp.stack(basis)

def thought_dim(n=None):
    global THOUGHT_DIM
    if n is not None:
        THOUGHT_DIM = n
    return THOUGHT_DIM

class Thought:

    def __init__(self, t=None):
        self._t = t

    def think(self):
        # initialize lazily
        if self._t is None:
            self._t = new_thought()
        return self._t

    def rethink(self, t):
        self._t = t
        return self

    def __repr__(self):
        return f"{self.__class__.__name__}({self._t})"

    def __array__(self):
        return self.think()




#!/usr/bin/env python3

import pytest

from think import Object, Type
from think import Bool, Int, Str, Float
from think import BoolType, StrType, IntType, FloatType


def test_core_meta_object_attribute():
    assert Type.object  is type
    assert BoolType.object  is type
    assert StrType.object   is type
    assert IntType.object   is type
    assert FloatType.object is type


def test_core_meta_base_attribute():
    assert BoolType.base  is Bool
    assert StrType.base   is Str
    assert IntType.base   is Int
    assert FloatType.base is Float


def test_core_meta_type_is_python_base_meta():
    assert type(Type)       is type
    assert type(BoolType)   is type
    assert type(StrType)    is type
    assert type(IntType)    is type
    assert type(FloatType)  is type


def test_core_meta_not_equal_implies_types_not_equal():
    assert Type('Ticker') != StrType('Ticker')


def test_core_meta_not_equal_implies_objects_not_equal():
    assert Type('Ticker')('SPY') != StrType('Ticker')('SPY')


if False:
    # obscure __classcell__ error occurring
    # when we make a class Char(Str): pass
    # *twice* (but not once), and this is
    # happening iff we cache types, so don't
    # cache types for now.
    def test_core_meta_idempotence():
        assert StrType('Name')   is StrType('Name')
        assert BoolType('Alive') is BoolType('Alive')
        assert IntType('Age')    is IntType('Age')
        assert FloatType('Size') is FloatType('Size')




#!/usr/bin/env python3

__all__ = [
    'hybridmethod',
    'metamethod',
]

# random

import jax.random

class State:
    def __init__(self, key=42):
        self.key = jax.random.PRNGKey(key)
    def split(self, key):
        key, subkey = jax.random.split(key)
        return key, subkey
    def normal(self, shape):
        self.key, subkey = self.split(self.key)
        return jax.random.normal(subkey, shape)


# descriptors

class hybridmethod:

    """
        A descriptor that lets classes share code
        with their instances, ignoring similarly
        named methods in the metaclass.
    """

    def __init__(self, func):
        self.func = func

    def __get__(self, obj, cls):
        func = self.func
        if obj is not None:
            return func.__get__(obj)
        else:
            return func.__get__(cls)


class metamethod:

    """
        In python,
        the type of object is type,
        and the parent of type is object.

        This descriptor gives us the same behavior,
        so that types have the same method resolution
        order as instances when we call methods on them.
    """

    def __init__(self, func):
        self.func = func
        self.name = func.__name__

    def __get__(self, obj, cls):
        if obj is not None:
            return self.func.__get__(obj)
        meta = cls.__class__
        try:
            func = getattr(meta, self.name)
        except AttributeError:
            func = self.func
        finally:
            return func.__get__(cls)




#!/usr/bin/env python3

import pytest

from think import Bool, Int, Str, Float
from think import BoolType, StrType, IntType, FloatType

Alive   = BoolType('Alive')
Name    = StrType('Name')
Age     = IntType('Age')
Weight  = FloatType('Weight')

def test_metacall_derived_bool_type_doesnt_accept_general_objects():
    with pytest.raises(TypeError):
        Alive(object())

def test_metacall_derived_str_type_doesnt_accept_general_objects():
    with pytest.raises(TypeError):
        Name(object())

def test_metacall_derived_int_type_doesnt_accept_general_objects():
    with pytest.raises(TypeError):
        Age(object())

def test_metacall_derived_float_type_doesnt_accept_general_objects():
    with pytest.raises(TypeError):
        Weight(object())


class Alive(Bool):
    pass
class Name(Str):
    pass
class Age(Int):
    pass
class Weight(Float):
    pass

def test_class_derived_bool_type_doesnt_accept_general_objects():
    with pytest.raises(TypeError):
        Alive(object())

def test_class_derived_str_type_doesnt_accept_general_objects():
    with pytest.raises(TypeError):
        Name(object())

def test_class_derived_int_type_doesnt_accept_general_objects():
    with pytest.raises(TypeError):
        Age(object())

def test_class_derived_float_type_doesnt_accept_general_objects():
    with pytest.raises(TypeError):
        Weight(object())




#!/usr/bin/env python3

__all__ = [
    'Add',
    'Sub',
    'Mul',
    'Div',
]

from think import slow
import think

class BinaryOperation:
    def __init__(self, a, b):
        self.a = a
        self.b = b
    def __repr__(self):
        return f"{self.a.object} {self.op} {self.b.object}"

class Add(BinaryOperation):
    op = '+'
    def think(self):
        return slow.mix([self.a.think(), +self.b.think()])

class Sub(BinaryOperation):
    op = '-'
    def think(self):
        return slow.mix([self.a.think(), -self.b.think()])

class ScalarOperation:
    def __init__(self, a, b):
        if isinstance(a, think.Object):
            assert not isinstance(b, think.Object)
            object = a
            scalar = b
        elif isinstance(b, think.Object):
            assert not isinstance(a, think.Object)
            object = b
            scalar = a
        else:
            raise TypeError(a, b)
        self.o = object
        self.s = scalar

    def __repr__(self):
        return f"{self.s} {self.op} {self.o.object}"

class Mul(ScalarOperation):
    op = '*'
    def think(self):
        return self.o.think() * self.s

class Div(ScalarOperation):
    op = '/'
    def think(self):
        return self.o.think() / self.s




#!/usr/bin/env python3

"""
    Thought geometry: JIT functions
"""

__all__ = [
    # unary functions on vectors
    'norm',
    'norm_l1',
    'norm_l2',
    'unit',
    'unit_l1',
    'unit_l2',
    'to_row',
    'to_col',

    # binary functions on vectors
    'dot',
    'proj',
    'dist',
    'cos',
    'breed',

    # unary functions of a hard coded plural argument
    'mean',
    'mix',

    # binary functions with at least one hard coded plural argument
    'coordinates',
    'expand',
    'split',
    'project',
    'reject',
    'explained',
    'unexplained',
    'pre_attention_l1',
    'pre_attention_l2',
    'pre_attention_sm',
    'attention_l1',
    'attention_l2',
    'attention_sm',
    'dots',

    # ternary functions with at least one hard coded plural argument
    'setattr',
    'mixattr',
]

import jax
import jax.numpy as jnp
from jax import jit
from jax.nn import softmax

##################################
### UNARY FUNCTIONS OF VECTORS ###
##################################

@jit
def norm(t):
    return jnp.sqrt(dot(t, t))

@jit
def norm_l1(t):
    return jnp.sum(t)

@jit
def norm_l2(t):
    return jnp.sqrt(dot(t, t))

@jit
def unit(t):
    return t/norm(t)

@jit
def unit_l1(t):
    return t/norm_l1(t)

@jit
def unit_l2(t):
    return t/norm_l2(t)

@jit
def to_row(t):
    assert t.ndim == 1
    return jnp.expand_dims(t, axis=0)

@jit
def to_col(t):
    assert t.ndim == 1
    return jnp.expand_dims(t, axis=-1)

###################################
### BINARY FUNCTIONS OF VECTORS ###
###################################

@jit
def dot(a, b):
    return (a*b).sum()

@jit
def proj(a, b):
    u = unit(b)
    return dot(a, u)*u

@jit
def rej(a, b):
    return a - proj(a, b)

@jit
def dist(a, b):
    return norm(a - b)

@jit
def cos(a, b):
    return dot(unit(a), unit(b))

@jit
def breed(a, b):
    return jnp.add(a, b)/jnp.sqrt(2)


#######################################################
### UNARY FUNCTIONS OF A HARD CODED PLURAL ARGUMENT ###
#######################################################

@jit
def mean(ts):
    return jnp.mean(ts, axis=0)

@jit
def mix(ts):
    n = ts.shape[0]
    return jnp.sum(ts, axis=0)/jnp.sqrt(n)


#####################################################################
### BINARY FUNCTIONS WITH AT LEAST ONE HARD CODED PLURAL ARGUMENT ###
#####################################################################

@jit
def coordinates(ts, t):
    coefs, err, rank, svals = jnp.linalg.lstsq(ts.T, to_col(t))
    return coefs

@jit
def expand(ts, t):
    ws = coordinates(ts, t)
    return ws*ts

@jit
def split(ts, t):
    basis = expand(ts, t)
    inner = jnp.sum(basis, axis=0)
    outer = t - inner
    return jnp.stack([inner, outer], axis=0)

@jit
def project(ts, t):
    basis = expand(ts, t)
    inner = jnp.sum(basis, axis=0)
    return inner

@jit
def reject(ts, t):
    return t - project(ts, t)

@jit
def explained(ts, t):
    i = project(ts, t)
    return (norm(i)/norm(t))**2

@jit
def unexplained(ts, t):
    o = reject(ts, t)
    return (norm(o)/norm(t))**2

@jit
def pre_attention_l1(ts, t):
    q = project(ts, t)
    dots = jnp.dot(ts, q)
    sims = dots/norm_l1(dots)
    return sims

@jit
def pre_attention_l2(ts, t):
    q = project(ts, t)
    dots = jnp.dot(ts, q)
    sims = dots/norm_l2(dots)
    return sims

@jit
def pre_attention_sm(ts, t):
    q = project(ts, t)
    dots = jnp.dot(ts, q)
    sims = softmax(dots)
    return sims

@jit
def attention_l1(ts, t):
    sims = pre_attention_l1(ts, t)
    return sims @ ts

@jit
def attention_l2(ts, t):
    sims = pre_attention_l2(ts, t)
    return sims @ ts

@jit
def attention_sm(ts, t):
    sims = pre_attention_sm(ts, t)
    return sims @ ts

@jit
def dots(ts, t):
    return jnp.dot(ts, t)

######################################################################
### TERNARY FUNCTIONS WITH AT LEAST ONE HARD CODED PLURAL ARGUMENT ###
######################################################################

@jit
def setattr(ts, t, v):
    t_in, t_out = split(ts, t)
    v_in, v_out = split(ts, v)
    return v_in + t_out

@jit
def mixattr(ts, t, v):
    t_in, t_out = split(ts, t)
    v_in, v_out = split(ts, v)
    v_frac = (norm(v_in)/norm(v))**2
    t_frac = (norm(t_in)/norm(t))**2
    weight = (1 - t_frac)*(v_frac)
    t_in2 = (weight)*v_in + (1 - weight)*t_in
    return t_in2 + t_out




#!/usr/bin/env python3

__all__ = [
    'set_log_level',
]

import os
import sys
import logging
import threading

def set_log_level(level, name='think'):
    logger = logging.getLogger(name)
    levelnum = getattr(logging, level.upper())
    logger.setLevel(levelnum)
    for handler in logger.handlers:
        if isinstance(handler, logging.FileHandler):
            continue
        handler.setLevel(levelnum)
    return logger


def get_logger(name):

    # create logger
    logger = logging.getLogger(name)

    # set level
    level = logging.INFO
    logger.setLevel(level)

    # create formatter
    format = '%(asctime)s.%(msecs)03d:%(name)s:%(funcName)s:%(levelname)s:%(message)s'
    formatter = logging.Formatter(format, datefmt='%Y-%m-%d_%H:%M:%S')

    # create stream handler
    stream = threading.main_thread()._stderr
    #stream = sys.stdout
    sh = logging.StreamHandler(stream)
    sh.setLevel(level)
    sh.setFormatter(formatter)
    logger.addHandler(sh)

    # create file handler and set level to debug
    if dir := os.getenv("THINK"):
        os.makedirs(dir, exist_ok=True)
        file = f"{dir}/think.log"
        fh = logging.FileHandler(file)
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(formatter)
        logger.addHandler(fh)

    return logger


root = get_logger('think')



#!/usr/bin/env python3


__all__ = [
    'Len',
    'Sign',
    'Digit',
    'Bit',
    'Decimal',
    'Binary',
    'LossyDecimal',
]

import re
import jax.numpy as jnp

from think import Int

class Len(Int):
    pass

class Sign(Int):
    pass

class Numeral(Int):

    @classmethod
    def __object__(cls, m, *args, **kwds):
        return int(m)

    def __new__(cls, m, *args, **kwds):
        """ Ordinal basis implementation """
        n = getattr(cls, 'Item', 0)
        a = cls[n].think()
        b = cls[n+1].think()
        θ = cls.theta(m)
        t = a*jnp.cos(θ) + b*jnp.sin(θ)
        self = super().__new__(cls, m, *args, **kwds)
        return self

    @classmethod
    def theta(cls, m):
        M = len(cls.__instances__)
        return (m/M)*(jnp.pi/2)


class Digit(Numeral):
    __instances__ = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

class Bit(Numeral):
    __instances__ = [0, 1]


class Decimal(Int):
    def __init__(self, n):
        digits = str(n)
        for slot, digit in enumerate(reversed(digits)):
            self.set(Digit[slot], int(digit))
        self.set(Len, len(digits))


class Decimal(Int):
    def __init__(self, n):
        sign = +1 if n >= 0 else -1
        digits = str(abs(n))
        for slot, digit in enumerate(reversed(digits)):
            self.set(Digit[slot], int(digit))
        self.set(Len, len(digits))
        self.set(Sign, sign)


class Binary(Int):
    def __init__(self, n):
        digits = bin(n)[2:]
        for slot, bit in enumerate(reversed(digits)):
            self.set(Bit[slot], int(bit))


class LossyDecimal(Int):
    def __init__(self, n, num_digits=1):
        sign = +1 if n >= 0 else -1
        digits = str(abs(n))
        for slot, digit in enumerate(digits):
            if slot < num_digits:
                print(f'self.set(Digit[{slot}], {digit})')
                self.set(Digit[slot], int(digit))
        self.set(Len, len(digits))
        self.set(Sign, sign)



#!/usr/bin/env python3

# core.py

"""
    All2Vec core.

    The python type system.

    This time it's differentiable.
"""

__all__ = [
    'Object',
    'Type',
    'IsInstance',
]

import abc
import types
import logging
import numbers
import logging
import builtins
import itertools
import jax.numpy as jnp

import think
from think import fast
from think import slow
from think import Thought, new_thought
from think.internals import hybridmethod, metamethod
from think.ops import Add, Sub, Mul, Div
from think import gradients
from think.pretty import colors


OBJECTS = {}
TYPES   = {}

# all 4 combinations of these two boolean variables
# should produce identical outcomes if the system is
# consistent.
DEFINE_OBJECTS_USING_CLASSES = True
DEFINE_TYPES_USING_CLASSES = True

# if this is set to True, then all types allow None as a value
ALL_TYPES_NULLABLE = True

logger = logging.getLogger(__name__)


# hashability
class thinklist(list):
    def __hash__(self):
        return hash(tuple(self))

class thinkdict(dict):
    def __hash__(self):
        return hash(tuple(self.items()))

class thinkset(set):
    def __hash__(self):
        return hash(tuple(self))

TYPE_PROMOTIONS = {
    list: thinklist,
    set: thinkset,
    dict: thinkdict,
}

def sort_dict(d):
    return dict(sorted(d.items(), key=lambda pair: pair[1], reverse=True))

class Type(type):

    """ A better way to implement Type. """

    object = type

    def __new__(meta, name, bases=(), dict=None, **kwds):

        # dict
        dict = {} if dict is None else dict
        dict |= kwds

        # bases
        if not bases:
            if 'base' in dict:
                bases = (dict['base'],)
            elif hasattr(meta, 'base'):
                bases = (meta.base,)
            elif hasattr(meta, 'firstborn'):
                bases = (meta.firstborn,)
            else:
                assert name == 'Object'
        elif isinstance(bases, type):
            bases = (bases,)
        else:
            for base in bases:
                assert isinstance(base, type)

        # upgrade methods to metamethods
        for k,v in dict.items():
            if k.startswith('__') and k.endswith('__'):
                continue
            if isinstance(v, types.FunctionType):
                dict[k] = metamethod(v)
                logger.debug(f"{name}.{k} upgraded to metamethod")

        # create the class
        cls = super().__new__(meta, name, bases, dict)

        if not hasattr(meta, 'firstborn'):
            meta.firstborn = cls

        cls.name = name
        cls.bases = [base for base in cls.mro() if hasattr(base, 'think')]
        cls.attrs = {}
        cls.thought = Thought()
        cls.kwds = kwds
        cls.subs = []
        cls.contexts = {}

        # list, set, dict:
        if cls.object in TYPE_PROMOTIONS and not hasattr(cls, '__object__'):
            cls.__object__ = TYPE_PROMOTIONS[cls.object]

        if hasattr(cls, '__create__'):
            assert isinstance(cls.__create__, (tuple, list, set))
            for o in cls.__create__:
                assert not isinstance(o, Object)
                cls(o)

        return cls

    def __init__(cls, name, bases=(), dict=None, **kwds):
        super().__init__(name, bases, dict)
        cls.memory = {}
        return None

    def __call__(cls, arg, *args, **kwds):

        if hasattr(cls, '__object__'):
            object = cls.__object__(arg, *args, **kwds)
        else:
            object = arg

        try:    hash(object)
        except: hashable = False
        else:   hashable = True

        if hashable:
            if object in cls.memory:
                return cls.memory[object]

        self = cls.__new__(cls, object, *args, **kwds)

        self.__raw__ = arg
        if isinstance(self, cls):
            cls.__init__(self, object, *args, **kwds)

        if hashable:
            cls.memory[object] = self

        if cls.primary:
            for base in cls.bases:
                if base is Object: continue
                self.set(IsInstance[base], True)
        return self


    def mro(cls):
        return type.mro(cls)

    def __repr__(cls):
        return f"{cls.__qualname__}"

    __module__ = None
    def __init_subclass__(meta):
        meta.__module__ = None

    def think(cls):
        return cls.thought.think()

    def rethink(cls, t):
        cls.thought.rethink(t)
        return cls

    if False:
        def __dir__(cls):
            # for normal object, dir contains keys from:
            # * self.__dict__
            # * cls.__dict__ for cls in self.__class__.mro()
            # so if we want classes to behave like normal objects w.r.t. tab completion,
            # the __dir__ for a class produced by a metaclass should be:
            self_dir  = [object.__dir__(cls)]
            base_dirs = [object.__dir__(base) for base in cls.mro()]
            all_dirs  = self_dir + base_dirs
            dir = sum(all_dirs, [])
            dir = sorted(set(dir))
            return dir

    ######################################
    ### plumbing for getattr / setattr ###
    ######################################

    def instances(cls):
        instances = {}
        instances |= cls.memory
        for base in cls.subs:
            instances |= base.memory
        return instances

    def similarities(cls, object):
        sims = cls.attention(object)
        keys = list(cls.memory.keys())
        return sort_dict(dict(zip(keys, sims)))

    def invert(cls, object):
        sims = cls.attention(object)
        idx  = int(jnp.argmax(sims))
        key  = list(cls.memory.keys())[idx]
        return key

    def project(cls, object):
        # HOLY FUCK THIS IS WAY BETTER FOR NON-ORTHOGONAL GENERALIZED BASES,
        # DON'T CHANGE IT, IT ABSOLUTELY CRUSHED THE ALTERNATIVES FOR THE
        # IMPORTANT (VERY-NON-ORTHOGONAL) CASE OF ORDINAL BASES.
        sims = cls.attention(object)
        idx  = int(jnp.argmax(sims))
        val  = list(cls.memory.values())[idx]
        return val

    def attention(cls, object):
        return slow.pre_attention_l1(cls, object)

    def __array__(cls):
        if len(cls.memory) == 0:
            raise Exception(f"Attribute {cls!r} doesn't yet have any instances.")
        vects = [o.__array__() for o in cls.memory.values()]
        return jnp.stack(vects, axis=0)

    ##########################
    ### self training code ###
    ##########################

    def perfect(cls):
        for name, self in cls.instances().items():
            if self.score() < 1.0:
                return False
        return True

    def learn(cls, **kwds):
        kwds.setdefault('threshold', 1.0)
        while not cls.perfect():
            for name, self in cls.instances().items():
                gradients.learn_until_score(self, **kwds)
        if cls is not Object:
            print(colors.white(f"{cls.name} is now perfect ✨"))
        else:
            print(colors.white(f"The system is now perfect ✨"))

    def learn_until_loss(cls, *args, **kwds):
        for name, self in cls.instances().items():
            gradients.learn_until_loss(self, *args, **kwds)

    def learn_until_score(cls, *args, **kwds):
        for name, self in cls.instances().items():
            gradients.learn_until_score(self, *args, **kwds)

    ###############################################
    ### context classes and __getitem__ support ###
    ###############################################

    primary = True


class Object(metaclass=Type):

    object = object

    def __new__(cls, object=None, t=None, **kwds):

        if ALL_TYPES_NULLABLE:
            if not isinstance(object, cls.object) and object is not None:
                raise TypeError(f"object {object!r} is not of type {cls.object}")
        else:
            if not isinstance(object, cls.object):
                raise TypeError(f"object {object!r} is not of type {cls.object}")

        self = builtins.object.__new__(cls)
        self.object  = object
        self.attrs   = {}
        self.thought = Thought(t)
        self.type    = type(cls)
        return self

    def __init__(self, *args, **kwds):
        pass

    def __init_subclass__(cls, **kwds):
        super().__init_subclass__()
        for base in cls.bases:
            base.subs.append(cls)

    def think(self):
        return self.thought.think()

    def rethink(self, t):
        self.thought.rethink(t)
        return self

    def bits(self):
        bits = 0
        for attr in self.attrs:
            bits += jnp.log2(len(attr.memory))
        return bits.item()

    def setfeel(self, attr, value):
        value = attr.ensure_thinkable(value)
        t = slow.setattr(attr, self, value)
        self.rethink(t)
        return self

    def setknow(self, attr, value):
        value = attr.ensure_thinkable(value)
        self.attrs[attr] = value
        return self

    def set(self, attr, value):
        value = attr.ensure_thinkable(value)
        # without the setfeel line, gradients are responsible for everything,
        # and the system is MUCH less able to learn quickly and sometimes
        # doesn't even converge.
        self.setfeel(attr, value)
        self.setknow(attr, value)
        return self

    def getfeel(self, attr, hard=False):
        thought = attr.project(self)
        return thought if not hard else attr.invert(thought)

    def getknow(self, attr, hard=False):
        thought = self.attrs.get(attr, Object(None))
        return thought if not hard else attr.invert(thought)

    def get(self, attr, how='feel', hard=True):
        if how == 'feel':
            return self.getfeel(attr, hard=hard)
        elif how == 'know':
            return self.getknow(attr, hard=hard)
        else:
            raise ValueError(f"how: {how!r}")

    def Get(self, attr):
        return attr(self.getfeel(attr, hard=True))

    @metamethod
    def __array__(self):
        return self.think()

    def __eq__(self, other):
        if isinstance(other, Object):
            return type(self) == type(other) and self.object == other.object
        else:
            return type(self.object) == type(other) and self.object == other

    def __repr__(self):
        return colors.blue(f"{self.object}")

    def __str__(self):
        return f"{self.__class__.__name__}({self.object})"

    def __bool__(self):
        return self.object is not None

    def __add__(self, other):
        return Add(self, other)

    def __sub__(self, other):
        return Sub(self, other)

    def __mul__(self, other):
        return Mul(self, other)

    def __truediv__(self, other):
        return Div(self, other)

    def unwrap(self):
        while hasattr(self, 'object'):
            self = self.object
        return self

    @classmethod
    def ensure_thinkable(cls, value):
        if isinstance(value, cls):
            return value
        elif isinstance(value, Object):
            cls(value.unwrap())
            return value
        else:
            return cls(value)

    @classmethod
    def ensure_unthinkable(cls, object):
        if isinstance(value, Object):
            return cls.unwrap(object)
        return object

    def reset_wrong(self):
        import random
        wrong = []
        items = list(self.attrs.items())
        random.shuffle(items)
        for attr, value in items:
            feel = self.get(attr)
            know = value.object
            if feel != know:
                wrong.append([attr, value])
        for attr, value in wrong:
            self.setfeel(attr, value)

    def score(self, reset_wrong=False):
        if not self.attrs:
            return 1.0
        right = []
        wrong = []
        for attr, value in self.attrs.items():
            feel = self.get(attr)
            know = value.object
            if feel == know:
                right.append([attr, value])
            else:
                wrong.append([attr, value])

        # compute score before resetting anything, so it's pure
        score = len(right)/len(self.attrs)

        # now reset the wrong ones if asked
        if reset_wrong:
            for attr, value in wrong:
                self.setfeel(attr, value)
        return score


    learn_until_loss  = gradients.learn_until_loss
    learn_until_score = gradients.learn_until_score

    @metamethod
    def perfect(self):
        return self.score() == 1.0

    @metamethod
    def learn(self):
        gradients.learn_until_score(self, threshold=1.0)
        print(colors.white(f"{self} is now perfect ✨"))

    def __class_getitem__(cls, item):
        #cls = cls.contextfree()
        try: return cls.contexts[item]
        except KeyError: pass
        name = f"{cls.__name__}[{item}]"
        sub = Type(name, cls, primary=False, Item=item)
        cls.contexts[item] = sub
        return sub


# In python:
#
# 1. type's type is type
# 2. type's base is object
# 3. object's base is object
# 4. object's type is type
#
# (see 53:30 to 54:00 here: https://youtu.be/uOzdG3lwcB4?t=3209)
#
# tl;dr: the solution to infinite regress is self reference.
Type.type = Type
Object.type = Type
Type.base = Object
Object.base = Object


# objects.py

__all__ += [
    'Bool',
    'Str',
    'Int',
    'Float',
    'Bytes',
    'Complex',
]

class IsInstance(Object):
    object = bool
IsInstance(True)
IsInstance(False)


if DEFINE_OBJECTS_USING_CLASSES:
    class Str(Object):
        object = str

    class Int(Object):
        object = int

    class Bool(Int):
        # bool.mro() is [bool, int, object]
        object = bool

    class Bytes(Object):
        object = bytes

    class Float(Object):
        object = float

    class Complex(Object):
        object = complex

else:
    Str     = Type('Str',     Object, object=str)
    Int     = Type('Int',     Object, object=int)
    Bool    = Type('Bool',    Int,    object=bool)
    Bytes   = Type('Bytes',   Object, object=bytes)
    Float   = Type('Float',   Object, object=float)
    Complex = Type('Complex', Object, object=complex)


# types.py

__all__ += [
    'BoolType',
    'StrType',
    'IntType',
    'FloatType',
    'BytesType',
    'ComplexType',
]

if DEFINE_TYPES_USING_CLASSES:
    class BoolType(Type):
        base = Bool

    class StrType(Type):
        base = Str

    class IntType(Type):
        base = Int

    class FloatType(Type):
        base = Float

    class BytesType(Type):
        base = Bytes

    class ComplexType(Type):
        base = Complex

else:
    BoolType    = type('BoolType', (Type,), {'base':Bool})
    StrType     = type('StrType', (Type,), {'base':Str})
    IntType     = type('IntType', (Type,), {'base':Int})
    FloatType   = type('FloatType', (Type,), {'base':Float})
    BytesType   = type('BytesType', (Type,), {'base':Bytes})
    ComplexType = type('ComplexType', (Type,), {'base':Complex})



__all__ += [
    'perfect',
    'learn',
]

perfect = Object.perfect
learn   = Object.learn

# thus ends the core
# all else is commentary




#!/usr/bin/env python

import pytest

import think
from think import Letter, Word, Sentence
from think import Year, Month, Day, Date


@pytest.mark.train
def test_training_a_system_to_perfection():

    # Example of training a system until it can
    # perfectly encode all the knowledge it has

    think.thought_dim(1024)

    a = Word('supercalifragilisticexpialidocious')
    b = Word('antidisestablishementarianism')
    c = Word('cupcake')
    d = Word('hotdog')
    for w in ('hey', 'there', 'babycakes', 'how', 'are', 'you', 'doing'):
        Word(w)

    s = Sentence("This is a fucking brain")
    t = Sentence("How cool is this")
    u = Sentence("That bitch crazy")
    v = Sentence("Yo mama so fat when she skips a meal the market drops")

    d1 = Date('1987-01-14')
    d2 = Date('2021-07-25')
    d3 = Date('2022-05-06')

    D1 = Date("September 17th, 03")
    D2 = Date("April 11th")
    D3 = Date("February 22nd")

    think.learn()

    assert think.perfect()

    # what we get from a perfectly trained system

    for word in (a,b,c,d):
        for n, letter in enumerate(word.object):
            assert word.get(Letter[n]) == letter

    for sentence in (s,t,u,v):
        for n, word in enumerate(sentence.object):
            assert sentence.get(Word[n]) == word

    for sentence in (s,t,u,v):
        for n, word in enumerate(sentence.object):
            assert sentence.get(Word[n]) == word
            word = Word(word)
            for m, letter in enumerate(word.object):
                assert word.get(Letter[m]) == letter

    for date in (d1,d2,d3):
        year, month, day = date.object.split('-')
        assert date.get(Year) == year
        assert date.get(Month) == month
        assert date.get(Day) == day

    assert D1.get(Year)  == '2003'
    assert D1.get(Month) == '09'
    assert D1.get(Day)   == '17'

    assert D2.get(Month) == '04'
    assert D2.get(Day)   == '11'

    assert D3.get(Month) == '02'
    assert D3.get(Day)   == '22'




#!/usr/bin/env python3

__all__ = [
    'Year',
    'Month',
    'Day',
    'Weekday',
    'Date',
]

import datetime

from think import Digit, Digits, Str

class Year(Digits):

    class Millenium(Digit):
        ...
    class Century(Digit):
        ...
    class Decade(Digit):
        ...
    class Year(Digit):
        ...

    def __init__(self, year):
        if not year.isnumeric():
            raise TypeError(f"Not a year: {year!r}")
        if len(year) == 4:
            pass
        elif len(year) == 2:
            year = f"20{year}" if year[0] in '012' else f"19{year}"
        else:
            raise TypeError(year)
        m,c,d,y = year
        self.set(self.Millenium, m)
        self.set(self.Century, c)
        self.set(self.Decade, d)
        self.set(self.Year, y)


class Month(Digits):
    pass

class Day(Digits):
    pass

class Weekday(Str):
    INT_TO_NAME =  {1:'Monday', 2:'Tuesday', 3:'Wednesday',
                    4:'Thursday', 5:'Friday', 6:'Saturday',
                    7:'Sunday'}
    NAME_TO_INT = {v:k for k,v in INT_TO_NAME.items()}

    @classmethod
    def __object__(cls, arg):
        if isinstance(arg, int) and 1 <= arg <= 7:
            return cls.INT_TO_NAME[arg]
        elif isinstance(arg, str) and arg in '1234567':
            return cls.INT_TO_NAME[int(arg)]
        elif isinstance(arg, str) and arg in cls.NAME_TO_INT:
            return arg
        else:
            raise ValueError(arg)


class Date(Str):

    def __init__(self, date, set_year=True, set_month=True, set_day=True, set_weekday=True):
        data = self.parse(date)
        if (year := data.get('year')) and set_year:
            self.set(Year, year)
        if (month := data.get('month')) and set_month:
            self.set(Month, month)
        if (day := data.get('day')) and set_day:
            self.set(Day, day)
        if (weekday := data.get('weekday')) and set_weekday:
            weekday = weekday() if callable(weekday) else weekday
            self.set(Weekday, weekday)

    @classmethod
    def date_object_to_dict(cls, date):
        d = {k: getattr(date, k) for k in ('year', 'month', 'day', 'weekday')}
        if callable(d['weekday']): # for 2/3 methods, it's a function
            d['weekday'] = d['weekday']()+1
        d = {k: str(v) for k,v in d.items()}
        lengths = {'year': 4, 'month': 2, 'day': 2, 'weekday': 1}
        for k in d:
            d[k] = d[k].zfill(lengths[k])
        return d

    @classmethod
    def suspicious_year(cls, input, output):
        this_year = datetime.date.today().year
        return  output['year'] == this_year \
        and     this_year not in re.findall(r'\d{4}', input)

    @classmethod
    def parse_with_timestring(cls, input):
        import timestring
        date = timestring.Date(input)
        output = cls.date_object_to_dict(date)
        if cls.suspicious_year(input, output):
            # timestring gives the present year when no year is found
            output['year'] = None
            output['weekday'] = None
        return output

    @classmethod
    def parse_with_dateparser(cls, input):
        import dateparser
        date = dateparser.parse(input)
        output = cls.date_object_to_dict(date)
        if cls.suspicious_year(input, output):
            # dateparser gives the present year when no year is found
            output['year'] = None
            output['weekday'] = None
        return output

    @classmethod
    def parse_with_datetime_module(cls, input):
        date = datetime.date.fromisoformat(input)
        output = cls.date_object_to_dict(date)
        return output

    def parse(cls, input):
        try:
            return cls.parse_with_timestring(input)
        except:
            pass
        try:
            return cls.parse_with_dateparser(input)
        except:
            pass
        try:
            return cls.parse_with_datetime_module(input)
        except:
            pass
        raise ValueError(input)




#!/usr/bin/env python

import pytest

import think
from think import Letter, Word, Sentence


@pytest.mark.train
def test_word():
    [Word(w) for w in ('Hello', 'nice', 'to', 'meet', 'you', 'goodbye')]
    w = Word("Hello")
    w.learn()
    assert w.get(Letter[0]) == 'H'
    assert w.get(Letter[1]) == 'e'
    assert w.get(Letter[2]) == 'l'
    assert w.get(Letter[3]) == 'l'
    assert w.get(Letter[4]) == 'o'

@pytest.mark.train
def test_sentence():
    s = Sentence("This is so cool")
    think.learn()
    assert Sentence.Item is Word
    assert Sentence.__object__('spam and eggs') == ('spam', 'and', 'eggs')
    assert s.get(Sentence.Item[0]) == 'This'
    assert s.get(Word[0]) == 'This'
    assert s.object == ('This', 'is', 'so', 'cool')
    assert s.__raw__ == 'This is so cool'




#!/usr/bin/env python3

"""
    Gradient based learning.
"""

__all__ = [
]

import logging
import jax.numpy as jnp
from jax import vmap, value_and_grad
try:
    # jax optimizers are here in newer versions
    from jax.example_libraries import optimizers
except:
    # jax optimizers are here in older versions
    from jax.experimental import optimizers

from think import fast
from think import slow
logger = logging.getLogger(__name__)


##########################
### SELF TRAINING CODE ###
##########################

def loss_fn(t, As, vs):
    feels = jnp.stack([fast.attention_l1(A, t) for A in As], axis=0)
    knows = jnp.stack(vs, axis=0)
    losses = (feels - knows)**2
    loss = losses.sum()
    return loss


grad_fn_self = value_and_grad(loss_fn)
grad_fn_all  = value_and_grad(loss_fn, argnums=[0,1,2])


def get_data_for_self_training(self, only_train_wrong=False):
    As = []
    vs = []
    attrs = []
    values = []
    for attr, value in self.attrs.items():
        feel = self.get(attr)
        know = value.object
        if only_train_wrong and feel == know:
            continue
        attrs.append(attr)
        values.append(value)
        A = slow.to_array(attr)
        v = value.think()
        vs.append(v)
        As.append(A)
    return (attrs, As), (values, vs)


def learn_until_score(self, threshold=1.0, step_size=1e-2,
                      optimizer=None, steps_per_update=20,
                      only_train_wrong=True, max_steps=float('inf')):

    # don't yammer about contextual types
    LOG = print # if type(self).primary else logger.debug

    score = self.score()
    if score >= threshold:
        LOG(f"no training needed for {self!r}, knowledge already encoded {score:.2%}")
        return self

    steps = 0
    num_proj = 0
    num_grad = 0

    while steps < max_steps:
        # train with projections
        self.reset_wrong()
        score = self.score()
        num_proj += 1

        if score >= threshold:
            LOG(f"{self!r}: projection got it: (projs={num_proj} grads={num_grad})")
            return self

        # train with gradients
        if steps == 0:
            LOG(f"training needed for {self!r}, knowledge encoded {score:.2%}, "
                f"will now train until {threshold:.2%}")
            t = self.think()
            (attrs, As), (values, vs) = get_data_for_self_training(self, only_train_wrong)
            if optimizer is None:
                optimizer = optimizers.adam
            opt_init, opt_update, get_params = optimizer(step_size)
            opt_state = opt_init(t)

        for n in range(steps_per_update):
            loss, grads = grad_fn_self(t, As, vs)
            opt_state = opt_update(0, grads, opt_state)
            t = get_params(opt_state)
        num_grad += 1
        self.rethink(t)

        score = self.score()
        if score >= threshold:
            LOG(f"{self!r}: gradients got it: (projs={num_proj} grads={num_grad})")
            # note: these are still naive projections
            return self

        LOG(f"{self!r}: finished {steps} steps. {score:.2%} of knowledge encoded, "
            f"desired {threshold:.2%} (loss={loss}, projs={num_proj} grads={num_grad})")

        steps += 1
    LOG(f"{self!r}: reached max steps, moving on.")
    return self


def learn_until_loss(self, threshold, step_size=1e-3,
                     optimizer=None, steps_per_update=20,
                     only_train_wrong=False, max_steps=float('inf')):

    # no need to do anything for no-knowledge objects
    if not self.attrs:
        return self

    if optimizer is None:
        optimizer = optimizers.adam

    # don't yammer about contextual types
    LOG = print if type(self).primary else logger.debug

    opt_init, opt_update, get_params = optimizer(step_size)

    t = self.think()
    (attrs, As), (values, vs) = get_data_for_self_training(self, only_train_wrong)

    loss = loss_fn(t, As, vs)
    if loss <= threshold:
        LOG(f"no training needed for {self!r}, loss is already {loss}")
        return self

    LOG(f"training needed for {self!r}, loss is {loss}, will now train until {threshold}")

    opt_state = opt_init(t)
    steps = 0
    while steps < max_steps:
        for n in range(steps_per_update):
            loss, grads = grad_fn_self(t, As, vs)
            opt_state = opt_update(0, grads, opt_state)
            t = get_params(opt_state)
            steps += 1
            if loss < threshold:
                LOG(f"{self!r}: training finished, encoded knowledge to loss: {loss}")
        self.rethink(t)
        self.reset_wrong()
        LOG(f"{self!r}: encoded knowledge to loss: {loss}")
    LOG(f"{self!r}: reached max steps, moving on.")
    return self




#!/usr/bin/env python3

__all__ = [
    'Digits',
    'Bits',
    'Char',
    'Letter',
    'Word',
    'Sentence',
]

import string

from think import Bool, Str
from think import Digit, Bit
from think import Tuple


# Numbers

class Digits(Str):
    """
        Digits
        ======
        A base10 number implementation using Str.
        This type works the same as Decimal, but it
        takes a string rather than an int.
        In data type like Year, we often want the
        leading zeros in instances like 0042 to have
        a specific meaning.
    """
    def __init__(self, digits):
        for slot, digit in enumerate(reversed(digits)):
            self.set(Digit[slot], int(digit))


class Bits(Str):
    def __init__(self, bits):
        if 'b' in bits:
            assert bits[:2] == '0b'
            bits = bits[2:]
        for slot, bit in enumerate(reversed(bits)):
            self.set(Bit[slot], int(bit))


# Letters

class Char(Str):
    def __init__(self, chr):
        if len(chr) > 1:
            raise TypeError(f"Not a char: {chr!r}")


class Letter(Char):

    LETTERS = set(string.ascii_letters)

    def __init__(self, letter):
        if letter not in self.LETTERS:
            raise TypeError(f"Not a letter: {letter!r}")


class Word(Str):
    def __init__(self, word):
        for n, letter in enumerate(word):
            self.set(Letter[n], letter)


class Sentence(Tuple[Word]):

    @classmethod
    def __object__(cls, str):
        try:
            import nltk
            words = super().__object__(nltk.tokenize.word_tokenize(str))
        except:
            words = str.split(' ') # poor man's tokenizer, if we don't have nltk
        return tuple(words)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.__raw__})"




#!/usr/bin/env python

from think import Object, Type
from think import Bool, Int, Str, Float

def test_core_types_type_is_our_base_meta():
    assert type(Object) is Type
    assert type(Bool)   is Type
    assert type(Str)    is Type
    assert type(Int)    is Type
    assert type(Float)  is Type




#!/usr/bin/env python3

# the core
from .log import *
from .internals import *
from .ops import *
from .thoughts import *
from .core import *

# the standard library
from .integers import *
from .collections import *
from .strings import *
from .datetimes import *



#!/usr/bin/env python3

from think import fast
from think import slow

__all__ = [
    'norm',
    'norm_l1',
    'norm_l2',
    'unit',
    'unit_l1',
    'unit_l2',
    'to_row',
    'to_col',
    'dot',
    'proj',
    'dist',
    'cos',
    'breed',
    'mean',
    'mix',
    'coordinates',
    'expand',
    'split',
    'project',
    'reject',
    'explained',
    'unexplained',
    'pre_attention_l1',
    'pre_attention_l2',
    'pre_attention_sm',
    'attention_l1',
    'attention_l2',
    'attention_sm',
    'dots',
    'setattr',
    'mixattr'
]

def norm(obj):
    t = slow.to_vector(obj)
    return fast.norm(t)

def norm_l1(obj):
    t = slow.to_vector(obj)
    return fast.norm_l1(t)

def norm_l2(obj):
    t = slow.to_vector(obj)
    return fast.norm_l2(t)

def unit(obj):
    t = slow.to_vector(obj)
    return fast.unit(t)

def unit_l1(obj):
    t = slow.to_vector(obj)
    return fast.unit_l1(t)

def unit_l2(obj):
    t = slow.to_vector(obj)
    return fast.unit_l2(t)

def to_row(obj):
    t = slow.to_vector(obj)
    return fast.to_row(t)

def to_col(obj):
    t = slow.to_vector(obj)
    return fast.to_col(t)

def dot(obj1, obj2):
    a = slow.to_vector(obj1)
    b = slow.to_vector(obj2)
    return fast.dot(a, b)

def proj(obj1, obj2):
    a = slow.to_vector(obj1)
    b = slow.to_vector(obj2)
    return fast.proj(a, b)

def dist(obj1, obj2):
    a = slow.to_vector(obj1)
    b = slow.to_vector(obj2)
    return fast.dist(a, b)

def cos(obj1, obj2):
    a = slow.to_vector(obj1)
    b = slow.to_vector(obj2)
    return fast.cos(a, b)

def breed(obj1, obj2):
    a = slow.to_vector(obj1)
    b = slow.to_vector(obj2)
    return fast.breed(a, b)

def mean(objs):
    ts = slow.to_array(objs)
    return fast.mean(ts)

def mix(objs):
    ts = slow.to_array(objs)
    return fast.mix(ts)

def coordinates(objs, obj):
    ts = slow.to_array(objs)
    t = slow.to_vector(obj)
    return fast.coordinates(ts, t)

def expand(objs, obj):
    ts = slow.to_array(objs)
    t = slow.to_vector(obj)
    return fast.expand(ts, t)

def split(objs, obj):
    ts = slow.to_array(objs)
    t = slow.to_vector(obj)
    return fast.split(ts, t)

def project(objs, obj):
    ts = slow.to_array(objs)
    t = slow.to_vector(obj)
    return fast.project(ts, t)

def reject(objs, obj):
    ts = slow.to_array(objs)
    t = slow.to_vector(obj)
    return fast.reject(ts, t)

def explained(objs, obj):
    ts = slow.to_array(objs)
    t = slow.to_vector(obj)
    return fast.explained(ts, t)

def unexplained(objs, obj):
    ts = slow.to_array(objs)
    t = slow.to_vector(obj)
    return fast.unexplained(ts, t)

def pre_attention_l1(objs, obj):
    ts = slow.to_array(objs)
    t = slow.to_vector(obj)
    return fast.pre_attention_l1(ts, t)

def pre_attention_l2(objs, obj):
    ts = slow.to_array(objs)
    t = slow.to_vector(obj)
    return fast.pre_attention_l2(ts, t)

def pre_attention_sm(objs, obj):
    ts = slow.to_array(objs)
    t = slow.to_vector(obj)
    return fast.pre_attention_sm(ts, t)

def attention_l1(objs, obj):
    ts = slow.to_array(objs)
    t = slow.to_vector(obj)
    return fast.attention_l1(ts, t)

def attention_l2(objs, obj):
    ts = slow.to_array(objs)
    t = slow.to_vector(obj)
    return fast.attention_l2(ts, t)

def attention_sm(objs, obj):
    ts = slow.to_array(objs)
    t = slow.to_vector(obj)
    return fast.attention_sm(ts, t)

def dots(objs, obj):
    ts = slow.to_array(objs)
    t = slow.to_vector(obj)
    return fast.dots(ts, t)

def setattr(objs, obj, value):
    ts = slow.to_array(objs)
    t = slow.to_vector(obj)
    v = slow.to_vector(value)
    return fast.setattr(ts, t, v)

def mixattr(objs, obj, value):
    ts = slow.to_array(objs)
    t = slow.to_vector(obj)
    v = slow.to_vector(value)
    return fast.mixattr(ts, t, v)



#!/usr/bin/env python3

import pytest

from think import Thought, new_thought
from think.internals import hybridmethod

def run_think_methods(a, A, descriptor):

    B = None
    b = None

    class O:
        t = Thought(A)
        def __init__(self):
            self.t = Thought(a)

        @descriptor
        def think(self):
            return self.t.think()

    o = O()

    try:    b = o.think()
    except: pass

    try:    B = O.think()
    except: pass

    return b, B

def test_classmethod_doesnt_work():
    a, b = new_thought(), new_thought()
    A, B = run_think_methods(a, b, classmethod)
    assert (a is not A) or (b is not B)

def test_staticmethod_doesnt_work():
    a, b = new_thought(), new_thought()
    A, B = run_think_methods(a, b, classmethod)
    assert (a is not A) or (b is not B)

def test_instancemethod_doesnt_work():
    a, b = new_thought(), new_thought()
    A, B = run_think_methods(a, b, lambda x: x)
    assert (a is not A) or (b is not B)

def test_hybridmethod_works():
    a, b = new_thought(), new_thought()
    A, B = run_think_methods(a, b, hybridmethod)
    assert (a is A) and (b is B)




#!/usr/bin/env python3

""" Metaprogramming. """

__all__ = [
    'build',
]

import os
import inspect

indent = ' '*4

LIFTED_NAMES = {
    'a': 'obj1',
    'b': 'obj2',
    't': 'obj',
    'v': 'value',
    'ts': 'objs',
}

def get_promotions(jit_func):
    triples = []
    num_plurals = 0
    for name in inspect.signature(jit_func).parameters:
        if len(name) == 1:
            promotion = 'slow.to_vector'
        elif len(name) == 2:
            promotion = 'slow.to_array'
            num_plurals += 1
        elif len(name) > 2:
            promotion = None
        else:
            raise ValueError(jit_func)

        if num_plurals >= 2:
            raise NotImplementedError(
                f"JIT function with >= 2 plurals encountered. "
                f"Implement this in meta.py."
            )

        old_name = name
        new_name = LIFTED_NAMES.get(name, name)
        triple = (old_name, new_name, promotion)
        triples.append(triple)

    return triples


def to_polymorphic_code(jit_func):

    triples = get_promotions(jit_func)

    old_names, new_names, promotions = zip(*triples)
    body = []
    for old_name, new_name, promotion in triples:
        if old_name != new_name:
            assert promotion.startswith('slow')
            line = f"{old_name} = {promotion}({new_name})"
            body.append(line)
        else:
            assert promotion is None

    name = jit_func.__name__
    new_args = ', '.join(new_names)
    old_args = ', '.join(old_names)
    head = f"def {name}({new_args}):"
    tail = f"return fast.{name}({old_args})"
    code = f"\n{indent}".join([head, *body, tail])
    return code

def to_polymorphic_function(jit_func):
    code = to_polymorphic_code(jit_func)
    namespace = {}
    exec(code, globals(), namespace)
    return namespace[name]


def regenerate_module(fast, output_path):

    exports = []
    functions = []

    for name in fast.__all__:
        jit_func = getattr(fast, name)
        exports.append(f"{name!r}")
        code = to_polymorphic_code(jit_func)
        functions.append(code)

    import textwrap

    imports = textwrap.dedent(f"""
    #!/usr/bin/env python3

    from think import fast
    from think import slow
    """)

    exports = f",\n{indent}".join(exports)

    functions = f"\n\n".join(functions)

    module = textwrap.dedent("""
    {imports}

    __all__ = [
        {exports}
    ]

    {functions}
    """).strip().format(
        imports=imports.strip(),
        exports=exports.strip(),
        functions=functions.strip(),
    )
    module = f"{module}\n"

    with open(output_path, 'w') as fp:
        fp.write(module)

    return module

def modification_time(path):
    if os.path.exists(path):
        return os.stat(path).st_mtime
    else:
        return 0

def build(force=False):
    from think import fast
    from think import slow
    dirname = os.path.dirname(slow.__file__)
    fast_path = fast.__file__
    slow_path = os.path.join(dirname, 'poly.py')
    last_path = os.path.join(dirname, '.lastbuild')
    if force:
        os.system(f"touch {last_path!r}")
        return regenerate_module(fast, slow_path)
    if not os.path.exists(slow_path):
        os.system(f"touch {last_path!r}")
        return regenerate_module(fast, slow_path)
    fast_mtime = modification_time(fast_path)
    slow_mtime = modification_time(slow_path)
    last_mtime = modification_time(last_path)
    if (last_mtime < fast_mtime) or (last_mtime < slow_mtime):
        os.system(f"touch {last_path!r}")
        return regenerate_module(fast, slow_path)
    return "Nothing needed"

if __name__ == '__main__':
    module = build(force=True)
    print(module)




#!/usr/bin/env python3

""" The pretty module """

__all__ = [
    'colors',
]

# colors

class color(str):
    end = "\033[00m"
    def __init__(self, code):
        self.code = code
    def __call__(self, string):
        return f"{self.code}{string}{self.end}"

class colors:
    black   = color("\033[01;30m")
    red     = color("\033[01;31m")
    green   = color("\033[01;32m")
    yellow  = color("\033[01;33m")
    blue    = color("\033[01;34m")
    purple  = color("\033[01;35m")
    cyan    = color("\033[01;36m")
    gray    = color("\033[01;37m")
    white   = color("\033[01;39m")




#!/usr/bin/env python

from think import Digit, Decimal

decimals = [Decimal(n) for n in range(100)]

def test_decimal_training_for_instances():
    d = Decimal(78)
    d.learn()
    assert d.get(Digit[0]) == 8
    assert d.get(Digit[1]) == 7

    d = Decimal(42)
    d.learn()
    assert d.get(Digit[0]) == 2
    assert d.get(Digit[1]) == 4

def test_decimal_training_for_class():
    Decimal.learn()
    for n, N in Decimal.instances().items():
        for k, digit in enumerate(reversed(str(n))):
            assert N.get(Digit[k]) == int(digit)




#!/usr/bin/env python3

__all__ = [
    'Item',
    'Key',
    'Value',
    'List',
    'Tuple',
    'Set',
    'Dict',
    'After',
    'Before',
    'LinkedList',
    'DoublyLinkedList',
]

from think import Object
from think import Len

# List, Tuple, Set

class Item(Object):
    """ For objects contained within other objects,
        like the elements of a list, tuple, or set. """
    pass

class Sequence(Object):

    Item = Item

    def __init__(self, seq):
        for n in range(len(seq)):
            self.set(self.Item[n], seq[n])
        self.set(Len, len(seq))

    def __getitem__(self, n):
        return self.get(self.Item[n])

    def __len__(self):
        return self.get(Len)

    @classmethod
    def invert(cls, t):
        items = [cls.Item[n].invert(t) for n in reversed(range(Len.invert(t)))]
        return items


class List(Sequence):
    object = list

class Tuple(Sequence):
    object = tuple

class Set(Sequence):
    object = set


# Dict

class Key(Object):
    pass


class Value(Object):
    pass


class Dict(Sequence):

    object = dict
    Key   = Key
    Value = Value

    def __init__(self, dict):
        for n, (k,v) in enumerate(dict.items()):
            self.set(self.Item[k], v)
            self.set(self.Key[n], k)
            self.set(self.Value[n], v)
        self.set(Len, len(dict))


# Linked Lists

class After(Object):
    pass

class Before(Object):
    pass

class Head(Object):
    pass

class LinkedList(Object):
    object = list
    Head = Head
    After = After
    Before = Before

    def __init__(self, list):
        items = [*list, None]
        self.set(self.Head, items[0])
        for this, next in zip(items[:-1], items[+1:]):
            self.set(self.After[this], next)


class DoublyLinkedList(LinkedList):

    def __init__(self, list):
        self.set(self.Head, None)
        items = [None, *list, None]
        for this, next in zip(items[:-1], items[+1:]):
            self.set(self.After[this], next)
            self.set(self.Before[next], this)





#!/usr/bin/env python

import pytest

from think import Object, Type
from think import Bool, Int, Str, Float
from think import BoolType, StrType, IntType, FloatType

def test_attribute_object_for_core_objects():
    assert Object(None).object  is None
    assert Bool(True).object    is True
    assert Int(42).object       is 42
    assert Str('hi').object     == 'hi'
    assert Float(0.1).object    == 0.1

def test_attribute_object_for_core_types():
    assert Object.object is object
    assert Bool.object   is bool
    assert Str.object    is str
    assert Int.object    is int
    assert Float.object  is float

def test_attribute_object_for_core_meta():
    # Python has many built-in types, but only one built-in metaclass.
    # Since .object gives the corresponding *python* object of each
    # learnable object, the correct response to these questions is
    # the python builtin: 'type'
    assert Type.object      is type
    assert BoolType.object  is type
    assert StrType.object   is type
    assert IntType.object   is type
    assert FloatType.object is type




#!/usr/bin/env python

import pytest

import think
from think import List, Tuple, Set, Dict
from think import LinkedList, DoublyLinkedList, After, Before


@pytest.mark.train
def test_list_tuple_dict_training():

    t = Tuple(('a', 42, 3+2j))
    l = List(['a', 42, 3+2j])
    d = Dict({'name': 'dave', 'age': 42})

    think.learn()

    assert l.get(List.Item[0]) == 'a'
    assert l.get(List.Item[1]) == 42
    assert l.get(List.Item[2]) == 3+2j

    assert t.get(Tuple.Item[0]) == 'a'
    assert t.get(Tuple.Item[1]) == 42
    assert t.get(Tuple.Item[2]) == 3+2j

    assert d.get(Dict.Key[0]) == 'name'
    assert d.get(Dict.Value[0]) == 'dave'
    assert d.get(Dict.Item['name']) == 'dave'
    assert d.get(Dict.Key[1]) == 'age'
    assert d.get(Dict.Value[1]) == 42
    assert d.get(Dict.Item['age']) == 42


@pytest.mark.train
def test_linked_list_training():

    list = ['This', 'is', 'a', 'linked', 'list']
    ll = LinkedList(list)

    think.learn()

    this = ll.get(ll.Head)
    walk = []
    while this is not None:
        walk.append(this)
        this = ll.get(ll.After[this])
    assert walk == list


@pytest.mark.train
def test_doubly_linked_list_training():

    list = ['This', 'is', 'a', 'doubly', 'linked', 'list']
    ld = DoublyLinkedList(list)

    ld.learn() # different

    forward = []
    head = ld.get(ld.Head)
    this = ld.get(ld.After[head])
    while this is not None:
        forward.append(this)
        this = ld.get(ld.After[this])

    backward = []
    head = ld.get(ld.Head)
    this = ld.get(ld.Before[head])
    while this is not None:
        backward.append(this)
        this = ld.get(ld.Before[this])

    assert forward  == list
    assert backward == list[::-1]




#!/usr/bin/env python3

import pytest

from think import Object, Bool, Int, Str, Float

def test_bool_idempotence():
    assert Bool(True) is Bool(True)

def test_str_idempotence():
    assert Str('cake') is Str('cake')

def test_int_idempotence():
    assert Int(3) is Int(3)

def test_float_idempotence():
    assert Float(42.69) is Float(42.69)

def test_bool_well_typed():
    with pytest.raises(TypeError):
        Bool(object())

def test_str_well_typed():
    with pytest.raises(TypeError):
        Str(object())

def test_int_well_typed():
    with pytest.raises(TypeError):
        Int(object())

def test_float_well_typed():
    with pytest.raises(TypeError):
        Float(object())

def test_core_objects_type_is_core_type():
    assert type(Object(None))   is Object
    assert type(Bool(True))     is Bool
    assert type(Str('cake'))    is Str
    assert type(Int(42))        is Int
    assert type(Float(42.69))   is Float

def test_core_objects_with_distinct_python_objects_are_not_equal():
    assert Bool(True) != Bool(False)
    assert Str('cake') != Str('pie')
    assert Int(3) != Int(4)
    assert Float(42.69) != Float(69.42)




#!/usr/bin/env python3

import pytest

from think import Bool, Int, Str, Float
from think import BoolType, StrType, IntType, FloatType

Alive   = BoolType('Alive')
Name    = StrType('Name')
Age     = IntType('Age')
Weight  = FloatType('Weight')

def test_metacall_derived_type_idempotence():
    assert Alive(True) is Alive(True)
    assert Name('Dave') is Name('Dave')
    assert Age(42) is Age(42)
    assert Weight(98.6) is Weight(98.6)

def test_metacall_derived_type_nontriviality():
    assert Alive(True) != Alive(False)
    assert Name('Dave') != Name('Bob')
    assert Age(42) != Age(43)
    assert Weight(42.69) != Weight(69.42)

def test_metacall_derived_objects_types():
    assert type(Alive(True))   is Alive
    assert type(Name('Dave'))  is Name
    assert type(Age(42))       is Age
    assert type(Weight(42.69)) is Weight


class Alive(Bool):  pass
class Name(Str):    pass
class Age(Int):     pass
class Size(Float):  pass

def test_class_derived_type_idempotence():
    assert Alive(True) is Alive(True)
    assert Name('Dave') is Name('Dave')
    assert Age(42) is Age(42)
    assert Weight(98.6) is Weight(98.6)

def test_class_derived_type_nontriviality():
    assert Alive(True) != Alive(False)
    assert Name('Dave') != Name('Bob')
    assert Age(42) != Age(43)
    assert Weight(42.69) != Weight(69.42)

def test_class_derived_objects_types():
    assert type(Alive(True))   is Alive
    assert type(Name('Dave'))  is Name
    assert type(Age(42))       is Age
    assert type(Weight(42.69)) is Weight
